@@ -1,4 +1,5 @@
 {-# LANGUAGE TemplateHaskell #-}
+{-# OPTIONS_GHC -fno-warn-warnings-deprecations #-}
 module Yesod.Core
     ( -- * Type classes
       Yesod (..)
@@ -7,6 +8,8 @@ module Yesod.Core
       -- ** Breadcrumbs
     , YesodBreadcrumbs (..)
     , breadcrumbs
+      -- * Types
+    , Approot (..)
       -- * Utitlities
     , maybeAuthorized
     , widgetToPageContent
@@ -14,6 +17,7 @@ module Yesod.Core
     , defaultErrorHandler
       -- * Data types
     , AuthResult (..)
+    , unauthorizedI
       -- * Logging
     , LogLevel (..)
     , formatLogMessage
@@ -23,6 +27,10 @@ module Yesod.Core
     , logWarn
     , logError
     , logOther
+    -- * JS loaders
+    , loadJsYepnope
+    , ScriptLoadPosition (..)
+    , BottomOfHeadAsync
       -- * Misc
     , yesodVersion
     , yesodRender
@@ -33,7 +41,6 @@ module Yesod.Core
     , module Yesod.Request
     , module Yesod.Widget
     , module Yesod.Message
-    , module Yesod.Config
     ) where
 
 import Yesod.Internal.Core
@@ -43,17 +50,17 @@ import Yesod.Handler
 import Yesod.Request
 import Yesod.Widget
 import Yesod.Message
-import Yesod.Config
 
 import Language.Haskell.TH.Syntax
+import qualified Language.Haskell.TH.Syntax as TH
 import Data.Text (Text)
 
 logTH :: LogLevel -> Q Exp
 logTH level =
-    [|messageLoggerHandler $(qLocation >>= liftLoc) $(lift level)|]
+    [|messageLoggerHandler $(qLocation >>= liftLoc) $(TH.lift level)|]
   where
     liftLoc :: Loc -> Q Exp
-    liftLoc (Loc a b c d e) = [|Loc $(lift a) $(lift b) $(lift c) $(lift d) $(lift e)|]
+    liftLoc (Loc a b c d e) = [|Loc $(TH.lift a) $(TH.lift b) $(TH.lift c) $(TH.lift d) $(TH.lift e)|]
 
 -- | Generates a function that takes a 'Text' and logs a 'LevelDebug' message. Usage:
 --
@@ -76,3 +83,9 @@ logError = logTH LevelError
 -- > $(logOther "My new level") "This is a log message"
 logOther :: Text -> Q Exp
 logOther = logTH . LevelOther
+
+-- | Return an 'Unauthorized' value, with the given i18n message.
+unauthorizedI :: RenderMessage master msg => msg -> GHandler sub master AuthResult
+unauthorizedI msg =do
+    mr <- getMessageRender
+    return $ Unauthorized $ mr msg
