@@ -16,7 +16,7 @@
 -- License       : BSD3
 --
 -- Maintainer    : Michael Snoyman <michael@snoyman.com>
--- Stability     : unstable
+-- Stability     : stable
 -- Portability   : portable
 --
 -- Define Handler stuff.
@@ -24,11 +24,9 @@
 ---------------------------------------------------------
 module Yesod.Handler
     ( -- * Type families
-      Route
-    , YesodSubRoute (..)
+      YesodSubRoute (..)
       -- * Handler monad
     , GHandler
-    , GGHandler
       -- ** Read information from handler
     , getYesod
     , getYesodSub
@@ -41,11 +39,9 @@ module Yesod.Handler
     , runRequestBody
       -- * Special responses
       -- ** Redirecting
-    , RedirectType (..)
+    , RedirectUrl (..)
     , redirect
-    , redirectParams
-    , redirectString
-    , redirectText
+    , redirectWith
     , redirectToPost
       -- ** Errors
     , notFound
@@ -63,6 +59,7 @@ module Yesod.Handler
     , sendWaiResponse
       -- * Setting headers
     , setCookie
+    , getExpires
     , deleteCookie
     , setHeader
     , setLanguage
@@ -74,14 +71,14 @@ module Yesod.Handler
       -- * Session
     , SessionMap
     , lookupSession
+    , lookupSessionBS
     , getSession
     , setSession
+    , setSessionBS
     , deleteSession
       -- ** Ultimate destination
     , setUltDest
-    , setUltDestString
-    , setUltDestText
-    , setUltDest'
+    , setUltDestCurrent
     , setUltDestReferer
     , redirectUltDest
     , clearUltDest
@@ -95,7 +92,8 @@ module Yesod.Handler
     , hamletToRepHtml
       -- ** Misc
     , newIdent
-    , liftIOHandler
+      -- * Lifting
+    , MonadLift (..)
       -- * i18n
     , getMessageRender
       -- * Per-request caching
@@ -123,7 +121,7 @@ module Yesod.Handler
 import Prelude hiding (catch)
 import Yesod.Internal.Request
 import Yesod.Internal
-import Data.Time (UTCTime)
+import Data.Time (UTCTime, getCurrentTime, addUTCTime)
 
 import Control.Exception hiding (Handler, catch, finally)
 import Control.Applicative
@@ -131,13 +129,12 @@ import Control.Applicative
 import Control.Monad (liftM)
 
 import Control.Monad.IO.Class
-import Control.Monad.Trans.Class
-import Control.Monad.Trans.Reader
+import Control.Monad.Trans.Class (MonadTrans)
+import qualified Control.Monad.Trans.Class
 
 import System.IO
 import qualified Network.Wai as W
 import qualified Network.HTTP.Types as H
-import Control.Failure (Failure (failure))
 
 import Text.Hamlet
 import qualified Text.Blaze.Renderer.Text
@@ -148,8 +145,6 @@ import qualified Data.Text.Lazy as TL
 
 import qualified Data.Map as Map
 import qualified Data.ByteString as S
-import Data.ByteString (ByteString)
-import Data.Enumerator (Iteratee (..), run_, ($$))
 import Network.Wai.Parse (parseHttpAccept)
 
 import Yesod.Content
@@ -160,20 +155,22 @@ import qualified Network.Wai.Parse as NWP
 import Data.Monoid (mappend, mempty, Endo (..))
 import qualified Data.ByteString.Char8 as S8
 import Data.CaseInsensitive (CI)
+import qualified Data.CaseInsensitive as CI
 import Blaze.ByteString.Builder (toByteString)
 import Data.Text (Text)
 import Yesod.Message (RenderMessage (..))
 
 import Text.Blaze (toHtml, preEscapedText)
-import Yesod.Internal.TestApi (catchIter)
 
 import qualified Yesod.Internal.Cache as Cache
 import Yesod.Internal.Cache (mkCacheKey, CacheKey)
 import Data.Typeable (Typeable)
 import qualified Data.IORef as I
-
--- | The type-safe URLs associated with a site argument.
-type family Route a
+import Control.Monad.Trans.Resource
+import Control.Exception.Lifted (catch)
+import Control.Monad.Trans.Control
+import Control.Monad.Base
+import Yesod.Routes.Class
 
 class YesodSubRoute s y where
     fromSubRoute :: s -> y -> Route s -> Route y
@@ -206,22 +203,22 @@ handlerSubDataMaybe tm ts route hd = hd
     , handlerRoute = route
     }
 
-get :: MonadIO monad => GGHandler sub master monad GHState
+get :: GHandler sub master GHState
 get = do
     hd <- ask
     liftIO $ I.readIORef $ handlerState hd
 
-put :: MonadIO monad => GHState -> GGHandler sub master monad ()
+put :: GHState -> GHandler sub master ()
 put g = do
     hd <- ask
     liftIO $ I.writeIORef (handlerState hd) g
 
-modify :: MonadIO monad => (GHState -> GHState) -> GGHandler sub master monad ()
+modify :: (GHState -> GHState) -> GHandler sub master ()
 modify f = do
     hd <- ask
     liftIO $ I.atomicModifyIORef (handlerState hd) $ \g -> (f g, ())
 
-tell :: MonadIO monad => Endo [Header] -> GGHandler sub master monad ()
+tell :: Endo [Header] -> GHandler sub master ()
 tell hs = modify $ \g -> g { ghsHeaders = ghsHeaders g `mappend` hs }
 
 -- | Used internally for promoting subsite handler functions to master site
@@ -229,19 +226,19 @@ tell hs = modify $ \g -> g { ghsHeaders = ghsHeaders g `mappend` hs }
 toMasterHandler :: (Route sub -> Route master)
                 -> (master -> sub)
                 -> Route sub
-                -> GGHandler sub master mo a
-                -> GGHandler sub' master mo a
-toMasterHandler tm ts route = withReaderT (handlerSubData tm ts route)
+                -> GHandler sub master a
+                -> GHandler sub' master a
+toMasterHandler tm ts route = local (handlerSubData tm ts route)
 
-toMasterHandlerDyn :: Monad mo
-                   => (Route sub -> Route master)
-                   -> GGHandler sub' master mo sub
+-- | FIXME do we need this?
+toMasterHandlerDyn :: (Route sub -> Route master)
+                   -> GHandler sub' master sub
                    -> Route sub
-                   -> GGHandler sub master mo a
-                   -> GGHandler sub' master mo a
+                   -> GHandler sub master a
+                   -> GHandler sub' master a
 toMasterHandlerDyn tm getSub route h = do
     sub <- getSub
-    withReaderT (handlerSubData tm (const sub) route) h
+    local (handlerSubData tm (const sub) route) h
 
 class SubsiteGetter g m s | g -> s where
   runSubsiteGetter :: g -> m s
@@ -258,18 +255,15 @@ instance (anySub ~ anySub'
 toMasterHandlerMaybe :: (Route sub -> Route master)
                      -> (master -> sub)
                      -> Maybe (Route sub)
-                     -> GGHandler sub master mo a
-                     -> GGHandler sub' master mo a
-toMasterHandlerMaybe tm ts route = withReaderT (handlerSubDataMaybe tm ts route)
+                     -> GHandler sub master a
+                     -> GHandler sub' master a
+toMasterHandlerMaybe tm ts route = local (handlerSubDataMaybe tm ts route)
 
 -- | A generic handler monad, which can have a different subsite and master
--- site. This monad is a combination of 'ReaderT' for basic arguments, a
--- 'WriterT' for headers and session, and an 'MEitherT' monad for handling
--- special responses. It is declared as a newtype to make compiler errors more
--- readable.
-type GGHandler sub master = ReaderT (HandlerData sub master)
-
-type GHandler sub master = GGHandler sub master (Iteratee ByteString IO)
+-- site. We define a newtype for better error message.
+newtype GHandler sub master a = GHandler
+    { unGHandler :: HandlerData sub master -> ResourceT IO a
+    }
 
 data GHState = GHState
     { ghsSession :: SessionMap
@@ -279,7 +273,7 @@ data GHState = GHState
     , ghsHeaders :: Endo [Header]
     }
 
-type SessionMap = Map.Map Text Text
+type SessionMap = Map.Map Text S.ByteString
 
 -- | An extension of the basic WAI 'W.Application' datatype to provide extra
 -- features needed by Yesod. Users should never need to use this directly, as
@@ -290,7 +284,7 @@ newtype YesodApp = YesodApp
     -> Request
     -> [ContentType]
     -> SessionMap
-    -> Iteratee ByteString IO YesodAppResult
+    -> ResourceT IO YesodAppResult
     }
 
 data YesodAppResult
@@ -301,7 +295,7 @@ data HandlerContents =
       HCContent H.Status ChooseRep
     | HCError ErrorResponse
     | HCSendFile ContentType FilePath (Maybe W.FilePart) -- FIXME replace FilePath with opaque type from system-filepath?
-    | HCRedirect RedirectType Text
+    | HCRedirect H.Status Text
     | HCCreated Text
     | HCWai W.Response
     deriving Typeable
@@ -310,11 +304,11 @@ instance Show HandlerContents where
     show _ = "Cannot show a HandlerContents"
 instance Exception HandlerContents
 
-getRequest :: Monad mo => GGHandler s m mo Request
+getRequest :: GHandler s m Request
 getRequest = handlerRequest `liftM` ask
 
-instance MonadIO monad => Failure ErrorResponse (GGHandler sub master monad) where
-    failure = liftIO . throwIO . HCError
+hcError :: ErrorResponse -> GHandler sub master a
+hcError = liftIO . throwIO . HCError
 
 runRequestBody :: GHandler s m RequestBodyContents
 runRequestBody = do
@@ -327,44 +321,42 @@ runRequestBody = do
             put x { ghsRBC = Just rbc }
             return rbc
 
-rbHelper :: W.Request -> Iteratee ByteString IO RequestBodyContents
+rbHelper :: W.Request -> ResourceT IO RequestBodyContents
 rbHelper req =
-    (map fix1 *** map fix2) <$> iter
+    (map fix1 *** map fix2) <$> (NWP.parseRequestBody NWP.lbsBackEnd req)
   where
-    iter = NWP.parseRequestBody NWP.lbsSink req
     fix1 = go *** go
     fix2 (x, NWP.FileInfo a b c) =
         (go x, FileInfo (go a) (go b) c)
     go = decodeUtf8With lenientDecode
 
 -- | Get the sub application argument.
-getYesodSub :: Monad m => GGHandler sub master m sub
+getYesodSub :: GHandler sub master sub
 getYesodSub = handlerSub `liftM` ask
 
 -- | Get the master site appliation argument.
-getYesod :: Monad m => GGHandler sub master m master
+getYesod :: GHandler sub master master
 getYesod = handlerMaster `liftM` ask
 
 -- | Get the URL rendering function.
-getUrlRender :: Monad m => GGHandler sub master m (Route master -> Text)
+getUrlRender :: GHandler sub master (Route master -> Text)
 getUrlRender = do
     x <- handlerRender `liftM` ask
     return $ flip x []
 
 -- | The URL rendering function with query-string parameters.
 getUrlRenderParams
-    :: Monad m
-    => GGHandler sub master m (Route master -> [(Text, Text)] -> Text)
+    :: GHandler sub master (Route master -> [(Text, Text)] -> Text)
 getUrlRenderParams = handlerRender `liftM` ask
 
 -- | Get the route requested by the user. If this is a 404 response- where the
 -- user requested an invalid route- this function will return 'Nothing'.
-getCurrentRoute :: Monad m => GGHandler sub master m (Maybe (Route sub))
+getCurrentRoute :: GHandler sub master (Maybe (Route sub))
 getCurrentRoute = handlerRoute `liftM` ask
 
 -- | Get the function to promote a route for a subsite to a route for the
 -- master site.
-getRouteToMaster :: Monad m => GGHandler sub master m (Route sub -> Route master)
+getRouteToMaster :: GHandler sub master (Route sub -> Route master)
 getRouteToMaster = handlerToMaster `liftM` ask
 
 -- | Function used internally by Yesod in the process of converting a
@@ -377,7 +369,7 @@ runHandler :: HasReps c
            -> master
            -> sub
            -> YesodApp
-runHandler handler mrender sroute tomr ma sa =
+runHandler handler mrender sroute tomr master sub =
   YesodApp $ \eh rr cts initSession -> do
     let toErrorHandler e =
             case fromException e of
@@ -392,14 +384,14 @@ runHandler handler mrender sroute tomr ma sa =
         }
     let hd = HandlerData
             { handlerRequest = rr
-            , handlerSub = sa
-            , handlerMaster = ma
+            , handlerSub = sub
+            , handlerMaster = master
             , handlerRoute = sroute
             , handlerRender = mrender
             , handlerToMaster = tomr
             , handlerState = istate
             }
-    contents' <- catchIter (fmap Right $ runReaderT handler hd)
+    contents' <- catch (fmap Right $ unGHandler handler hd)
         (\e -> return $ Left $ maybe (HCError $ toErrorHandler e) id
                       $ fromException e)
     state <- liftIO $ I.readIORef istate
@@ -420,12 +412,12 @@ runHandler handler mrender sroute tomr ma sa =
             (ct, c) <- liftIO $ a cts
             return $ YARPlain status (appEndo headers []) ct c finalSession
         HCError e -> handleError e
-        HCRedirect rt loc -> do
+        HCRedirect status loc -> do
             let hs = Header "Location" (encodeUtf8 loc) : appEndo headers []
             return $ YARPlain
-                (getRedirectStatus rt) hs typePlain emptyContent
+                status hs typePlain emptyContent
                 finalSession
-        HCSendFile ct fp p -> catchIter
+        HCSendFile ct fp p -> catch
             (sendFile' ct fp p)
             (handleError . toErrorHandler)
         HCCreated loc -> do
@@ -449,22 +441,26 @@ safeEh er = YesodApp $ \_ _ _ session -> do
         session
 
 -- | Redirect to the given route.
-redirect :: MonadIO mo => RedirectType -> Route master -> GGHandler sub master mo a
-redirect rt url = redirectParams rt url []
+-- HTTP status code 303 for HTTP 1.1 clients and 302 for HTTP 1.0
+-- This is the appropriate choice for a get-following-post
+-- technique, which should be the usual use case.
+--
+-- If you want direct control of the final status code, or need a different
+-- status code, please use 'redirectWith'.
+redirect :: RedirectUrl master url => url -> GHandler sub master a
+redirect url = do
+    req <- waiRequest
+    let status =
+            if W.httpVersion req == H.http11
+                then H.status303
+                else H.status302
+    redirectWith status url
 
--- | Redirects to the given route with the associated query-string parameters.
-redirectParams :: MonadIO mo
-               => RedirectType -> Route master -> [(Text, Text)]
-               -> GGHandler sub master mo a
-redirectParams rt url params = do
-    r <- getUrlRenderParams
-    redirectString rt $ r url params
-
--- | Redirect to the given URL.
-redirectString, redirectText :: MonadIO mo => RedirectType -> Text -> GGHandler sub master mo a
-redirectText rt = liftIO . throwIO . HCRedirect rt
-redirectString = redirectText
-{-# DEPRECATED redirectString "Use redirectText instead" #-}
+-- | Redirect to the given URL with the specified status code.
+redirectWith :: RedirectUrl master url => H.Status -> url -> GHandler sub master a
+redirectWith status url = do
+    urlText <- toTextUrl url
+    liftIO $ throwIO $ HCRedirect status urlText
 
 ultDestKey :: Text
 ultDestKey = "_ULT"
@@ -473,38 +469,29 @@ ultDestKey = "_ULT"
 --
 -- An ultimate destination is stored in the user session and can be loaded
 -- later by 'redirectUltDest'.
-setUltDest :: MonadIO mo => Route master -> GGHandler sub master mo ()
-setUltDest dest = do
-    render <- getUrlRender
-    setUltDestString $ render dest
-
--- | Same as 'setUltDest', but use the given string.
-setUltDestText :: MonadIO mo => Text -> GGHandler sub master mo ()
-setUltDestText = setSession ultDestKey
-
-setUltDestString :: MonadIO mo => Text -> GGHandler sub master mo ()
-setUltDestString = setSession ultDestKey
-{-# DEPRECATED setUltDestString "Use setUltDestText instead" #-}
+setUltDest :: RedirectUrl master url => url -> GHandler sub master ()
+setUltDest url = do
+    urlText <- toTextUrl url
+    setSession ultDestKey urlText
 
 -- | Same as 'setUltDest', but uses the current page.
 --
 -- If this is a 404 handler, there is no current page, and then this call does
 -- nothing.
-setUltDest' :: MonadIO mo => GGHandler sub master mo ()
-setUltDest' = do
+setUltDestCurrent :: GHandler sub master ()
+setUltDestCurrent = do
     route <- getCurrentRoute
     case route of
         Nothing -> return ()
         Just r -> do
             tm <- getRouteToMaster
             gets' <- reqGetParams `liftM` handlerRequest `liftM` ask
-            render <- getUrlRenderParams
-            setUltDestString $ render (tm r) gets'
+            setUltDest (tm r, gets')
 
 -- | Sets the ultimate destination to the referer request header, if present.
 --
 -- This function will not overwrite an existing ultdest.
-setUltDestReferer :: MonadIO mo => GGHandler sub master mo ()
+setUltDestReferer :: GHandler sub master ()
 setUltDestReferer = do
     mdest <- lookupSession ultDestKey
     maybe
@@ -512,23 +499,25 @@ setUltDestReferer = do
         (const $ return ())
         mdest
   where
-    setUltDestBS = setUltDestText . T.pack . S8.unpack
+    setUltDestBS = setUltDest . T.pack . S8.unpack
 
 -- | Redirect to the ultimate destination in the user's session. Clear the
 -- value from the session.
 --
 -- The ultimate destination is set with 'setUltDest'.
-redirectUltDest :: MonadIO mo
-                => RedirectType
-                -> Route master -- ^ default destination if nothing in session
-                -> GGHandler sub master mo a
-redirectUltDest rt def = do
+--
+-- This function uses 'redirect', and thus will perform a temporary redirect to
+-- a GET request.
+redirectUltDest :: RedirectUrl master url
+                => url -- ^ default destination if nothing in session
+                -> GHandler sub master a
+redirectUltDest def = do
     mdest <- lookupSession ultDestKey
     deleteSession ultDestKey
-    maybe (redirect rt def) (redirectText rt) mdest
+    maybe (redirect def) redirect mdest
 
 -- | Remove a previously set ultimate destination. See 'setUltDest'.
-clearUltDest :: MonadIO mo => GGHandler sub master mo ()
+clearUltDest :: GHandler sub master ()
 clearUltDest = deleteSession ultDestKey
 
 msgKey :: Text
@@ -537,13 +526,13 @@ msgKey = "_MSG"
 -- | Sets a message in the user's session.
 --
 -- See 'getMessage'.
-setMessage :: MonadIO mo => Html -> GGHandler sub master mo ()
+setMessage :: Html -> GHandler sub master ()
 setMessage = setSession msgKey . T.concat . TL.toChunks . Text.Blaze.Renderer.Text.renderHtml
 
 -- | Sets a message in the user's session.
 --
 -- See 'getMessage'.
-setMessageI :: (RenderMessage y msg, MonadIO mo) => msg -> GGHandler sub y mo ()
+setMessageI :: (RenderMessage y msg) => msg -> GHandler sub y ()
 setMessageI msg = do
     mr <- getMessageRender
     setMessage $ toHtml $ mr msg
@@ -552,7 +541,7 @@ setMessageI msg = do
 -- variable.
 --
 -- See 'setMessage'.
-getMessage :: MonadIO mo => GGHandler sub master mo (Maybe Html)
+getMessage :: GHandler sub master (Maybe Html)
 getMessage = do
     mmsg <- liftM (fmap preEscapedText) $ lookupSession msgKey
     deleteSession msgKey
@@ -562,34 +551,33 @@ getMessage = do
 --
 -- For some backends, this is more efficient than reading in the file to
 -- memory, since they can optimize file sending via a system call to sendfile.
-sendFile :: MonadIO mo => ContentType -> FilePath -> GGHandler sub master mo a
+sendFile :: ContentType -> FilePath -> GHandler sub master a
 sendFile ct fp = liftIO . throwIO $ HCSendFile ct fp Nothing
 
 -- | Same as 'sendFile', but only sends part of a file.
-sendFilePart :: MonadIO mo
-             => ContentType
+sendFilePart :: ContentType
              -> FilePath
              -> Integer -- ^ offset
              -> Integer -- ^ count
-             -> GGHandler sub master mo a
+             -> GHandler sub master a
 sendFilePart ct fp off count =
     liftIO . throwIO $ HCSendFile ct fp $ Just $ W.FilePart off count
 
 -- | Bypass remaining handler code and output the given content with a 200
 -- status code.
-sendResponse :: (MonadIO mo, HasReps c) => c -> GGHandler sub master mo a
+sendResponse :: HasReps c => c -> GHandler sub master a
 sendResponse = liftIO . throwIO . HCContent H.status200
              . chooseRep
 
 -- | Bypass remaining handler code and output the given content with the given
 -- status code.
-sendResponseStatus :: (MonadIO mo, HasReps c) => H.Status -> c -> GGHandler s m mo a
+sendResponseStatus :: HasReps c => H.Status -> c -> GHandler s m a
 sendResponseStatus s = liftIO . throwIO . HCContent s
                      . chooseRep
 
 -- | Send a 201 "Created" response with the given route as the Location
 -- response header.
-sendResponseCreated :: MonadIO mo => Route m -> GGHandler s m mo a
+sendResponseCreated :: Route m -> GHandler s m a
 sendResponseCreated url = do
     r <- getUrlRender
     liftIO . throwIO $ HCCreated $ r url
@@ -599,105 +587,124 @@ sendResponseCreated url = do
 -- that you have already specified. This function short-circuits. It should be
 -- considered only for very specific needs. If you are not sure if you need it,
 -- you don't.
-sendWaiResponse :: MonadIO mo => W.Response -> GGHandler s m mo b
+sendWaiResponse :: W.Response -> GHandler s m b
 sendWaiResponse = liftIO . throwIO . HCWai
 
 -- | Return a 404 not found page. Also denotes no handler available.
-notFound :: Failure ErrorResponse m => m a
-notFound = failure NotFound
+notFound :: GHandler sub master a
+notFound = hcError NotFound
 
 -- | Return a 405 method not supported page.
-badMethod :: MonadIO mo => GGHandler s m mo a
+badMethod :: GHandler sub master a
 badMethod = do
     w <- waiRequest
-    failure $ BadMethod $ W.requestMethod w
+    hcError $ BadMethod $ W.requestMethod w
 
 -- | Return a 403 permission denied page.
-permissionDenied :: Failure ErrorResponse m => Text -> m a
-permissionDenied = failure . PermissionDenied
+permissionDenied :: Text -> GHandler sub master a
+permissionDenied = hcError . PermissionDenied
 
 -- | Return a 403 permission denied page.
-permissionDeniedI :: (RenderMessage y msg, MonadIO mo) => msg -> GGHandler s y mo a
+permissionDeniedI :: RenderMessage master msg => msg -> GHandler sub master a
 permissionDeniedI msg = do
     mr <- getMessageRender
     permissionDenied $ mr msg
 
 -- | Return a 400 invalid arguments page.
-invalidArgs :: Failure ErrorResponse m => [Text] -> m a
-invalidArgs = failure . InvalidArgs
+invalidArgs :: [Text] -> GHandler sub master a
+invalidArgs = hcError . InvalidArgs
 
 -- | Return a 400 invalid arguments page.
-invalidArgsI :: (RenderMessage y msg, MonadIO mo) => [msg] -> GGHandler s y mo a
+invalidArgsI :: RenderMessage y msg => [msg] -> GHandler s y a
 invalidArgsI msg = do
     mr <- getMessageRender
     invalidArgs $ map mr msg
 
 ------- Headers
 -- | Set the cookie on the client.
-setCookie :: MonadIO mo
-          => Int -- ^ minutes to timeout
-          -> H.Ascii -- ^ key
-          -> H.Ascii -- ^ value
-          -> GGHandler sub master mo ()
-setCookie a b = addHeader . AddCookie a b
+
+setCookie :: SetCookie
+          -> GHandler sub master ()
+setCookie = addHeader . AddCookie
+
+-- | Helper function for setCookieExpires value
+getExpires :: Int -- ^ minutes
+          -> IO UTCTime
+getExpires m = do
+    now <- liftIO getCurrentTime
+    return $ fromIntegral (m * 60) `addUTCTime` now
+
 
 -- | Unset the cookie on the client.
-deleteCookie :: MonadIO mo => H.Ascii -> GGHandler sub master mo ()
-deleteCookie = addHeader . DeleteCookie
+--
+-- Note: although the value used for key and path is 'Text', you should only
+-- use ASCII values to be HTTP compliant.
+deleteCookie :: Text -- ^ key 
+             -> Text -- ^ path
+             -> GHandler sub master ()
+deleteCookie a = addHeader . DeleteCookie (encodeUtf8 a) . encodeUtf8
+
 
 -- | Set the language in the user session. Will show up in 'languages' on the
 -- next request.
-setLanguage :: MonadIO mo => Text -> GGHandler sub master mo ()
+setLanguage :: Text -> GHandler sub master ()
 setLanguage = setSession langKey
 
 -- | Set an arbitrary response header.
-setHeader :: MonadIO mo
-          => CI H.Ascii -> H.Ascii -> GGHandler sub master mo ()
-setHeader a = addHeader . Header a
+--
+-- Note that, while the data type used here is 'Text', you must provide only
+-- ASCII value to be HTTP compliant.
+setHeader :: Text -> Text -> GHandler sub master ()
+setHeader a = addHeader . Header (encodeUtf8 a) . encodeUtf8
 
 -- | Set the Cache-Control header to indicate this response should be cached
 -- for the given number of seconds.
-cacheSeconds :: MonadIO mo => Int -> GGHandler s m mo ()
-cacheSeconds i = setHeader "Cache-Control" $ S8.pack $ concat
+cacheSeconds :: Int -> GHandler s m ()
+cacheSeconds i = setHeader "Cache-Control" $ T.concat
     [ "max-age="
-    , show i
+    , T.pack $ show i
     , ", public"
     ]
 
 -- | Set the Expires header to some date in 2037. In other words, this content
 -- is never (realistically) expired.
-neverExpires :: MonadIO mo => GGHandler s m mo ()
+neverExpires :: GHandler s m ()
 neverExpires = setHeader "Expires" "Thu, 31 Dec 2037 23:55:55 GMT"
 
 -- | Set an Expires header in the past, meaning this content should not be
 -- cached.
-alreadyExpired :: MonadIO mo => GGHandler s m mo ()
+alreadyExpired :: GHandler s m ()
 alreadyExpired = setHeader "Expires" "Thu, 01 Jan 1970 05:05:05 GMT"
 
 -- | Set an Expires header to the given date.
-expiresAt :: MonadIO mo => UTCTime -> GGHandler s m mo ()
-expiresAt = setHeader "Expires" . encodeUtf8 . formatRFC1123
+expiresAt :: UTCTime -> GHandler s m ()
+expiresAt = setHeader "Expires" . formatRFC1123
 
 -- | Set a variable in the user's session.
 --
 -- The session is handled by the clientsession package: it sets an encrypted
 -- and hashed cookie on the client. This ensures that all data is secure and
 -- not tampered with.
-setSession :: MonadIO mo
-           => Text -- ^ key
+setSession :: Text -- ^ key
            -> Text -- ^ value
-           -> GGHandler sub master mo ()
-setSession k = modify . modSession . Map.insert k
+           -> GHandler sub master ()
+setSession k = setSessionBS k . encodeUtf8
+
+-- | Same as 'setSession', but uses binary data for the value.
+setSessionBS :: Text
+             -> S.ByteString
+             -> GHandler sub master ()
+setSessionBS k = modify . modSession . Map.insert k
 
 -- | Unsets a session variable. See 'setSession'.
-deleteSession :: MonadIO mo => Text -> GGHandler sub master mo ()
+deleteSession :: Text -> GHandler sub master ()
 deleteSession = modify . modSession . Map.delete
 
 modSession :: (SessionMap -> SessionMap) -> GHState -> GHState
 modSession f x = x { ghsSession = f $ ghsSession x }
 
 -- | Internal use only, not to be confused with 'setHeader'.
-addHeader :: MonadIO mo => Header -> GGHandler sub master mo ()
+addHeader :: Header -> GHandler sub master ()
 addHeader = tell . Endo . (:)
 
 getStatus :: ErrorResponse -> H.Status
@@ -707,42 +714,56 @@ getStatus (InvalidArgs _) = H.status400
 getStatus (PermissionDenied _) = H.status403
 getStatus (BadMethod _) = H.status405
 
-getRedirectStatus :: RedirectType -> H.Status
-getRedirectStatus RedirectPermanent = H.status301
-getRedirectStatus RedirectTemporary = H.status302
-getRedirectStatus RedirectSeeOther = H.status303
+-- | Some value which can be turned into a URL for redirects.
+class RedirectUrl master a where
+    -- | Converts the value to the URL and a list of query-string parameters.
+    toTextUrl :: a -> GHandler sub master Text
 
--- | Different types of redirects.
-data RedirectType = RedirectPermanent
-                  | RedirectTemporary
-                  | RedirectSeeOther
-    deriving (Show, Eq)
+instance RedirectUrl master Text where
+    toTextUrl = return
 
-localNoCurrent :: Monad mo => GGHandler s m mo a -> GGHandler s m mo a
+instance RedirectUrl master String where
+    toTextUrl = toTextUrl . T.pack
+
+instance RedirectUrl master (Route master) where
+    toTextUrl url = do
+        r <- getUrlRender
+        return $ r url
+
+instance (key ~ Text, val ~ Text) => RedirectUrl master (Route master, [(key, val)]) where
+    toTextUrl (url, params) = do
+        r <- getUrlRenderParams
+        return $ r url params
+
+localNoCurrent :: GHandler s m a -> GHandler s m a
 localNoCurrent =
     local (\hd -> hd { handlerRoute = Nothing })
 
 -- | Lookup for session data.
-lookupSession :: MonadIO mo => Text -> GGHandler s m mo (Maybe Text)
-lookupSession n = do
+lookupSession :: Text -> GHandler s m (Maybe Text)
+lookupSession = (fmap . fmap) (decodeUtf8With lenientDecode) . lookupSessionBS
+
+-- | Lookup for session data in binary format.
+lookupSessionBS :: Text -> GHandler s m (Maybe S.ByteString)
+lookupSessionBS n = do
     m <- liftM ghsSession get
     return $ Map.lookup n m
 
 -- | Get all session variables.
-getSession :: MonadIO mo => GGHandler s m mo SessionMap
+getSession :: GHandler sub master SessionMap
 getSession = liftM ghsSession get
 
 handlerToYAR :: (HasReps a, HasReps b)
-             => m -- ^ master site foundation
-             -> s -- ^ sub site foundation
-             -> (Route s -> Route m)
-             -> (Route m -> [(Text, Text)] -> Text)
-             -> (ErrorResponse -> GHandler s m a)
+             => master -- ^ master site foundation
+             -> sub    -- ^ sub site foundation
+             -> (Route sub -> Route master)
+             -> (Route master -> [(Text, Text)] -> Text) -- route renderer
+             -> (ErrorResponse -> GHandler sub master a)
              -> Request
-             -> Maybe (Route s)
+             -> Maybe (Route sub)
              -> SessionMap
-             -> GHandler s m b
-             -> Iteratee ByteString IO YesodAppResult
+             -> GHandler sub master b
+             -> ResourceT IO YesodAppResult
 handlerToYAR y s toMasterRoute render errorHandler rr murl sessionMap h =
     unYesodApp ya eh' rr types sessionMap
   where
@@ -764,31 +785,12 @@ yarToResponse renderHeaders (YARPlain s hs ct c sessionFinal) =
             let hs' = maybe finalHeaders finalHeaders' mlen
              in W.ResponseBuilder s hs' b
         ContentFile fp p -> W.ResponseFile s finalHeaders fp p
-        ContentEnum e ->
-            W.ResponseEnumerator $ \iter -> run_ $ e $$ iter s finalHeaders
+        ContentSource body -> W.ResponseSource s finalHeaders body
   where
     finalHeaders = renderHeaders hs ct sessionFinal
     finalHeaders' len = ("Content-Length", S8.pack $ show len)
                       : finalHeaders
-    {-
-    getExpires m = fromIntegral (m * 60) `addUTCTime` now
-    sessionVal =
-        case key' of
-            Nothing -> B.empty
-            Just key'' -> encodeSession key'' exp' host
-                        $ Map.toList
-                        $ Map.insert nonceKey (reqNonce rr) sessionFinal
-    hs' =
-            case key' of
-                Nothing -> hs
-                Just _ -> AddCookie
-                        (clientSessionDuration y)
-                        sessionName
-                        (bsToChars sessionVal)
-                      : hs
-    hs'' = map (headerToPair getExpires) hs'
-    hs''' = ("Content-Type", charsToBs ct) : hs''
-    -}
+
 
 httpAccept :: W.Request -> [ContentType]
 httpAccept = parseHttpAccept
@@ -797,40 +799,28 @@ httpAccept = parseHttpAccept
            . W.requestHeaders
 
 -- | Convert Header to a key/value pair.
-headerToPair :: S.ByteString -- ^ cookie path
-             -> (Int -> UTCTime) -- ^ minutes -> expiration time
-             -> Header
+headerToPair :: Header
              -> (CI H.Ascii, H.Ascii)
-headerToPair cp getExpires (AddCookie minutes key value) =
-    ("Set-Cookie", toByteString $ renderSetCookie $ SetCookie
-        { setCookieName = key
-        , setCookieValue = value
-        , setCookiePath = Just cp
-        , setCookieExpires =
-            if minutes == 0
-                then Nothing
-                else Just $ getExpires minutes
-        , setCookieDomain = Nothing
-        , setCookieHttpOnly = True
-        })
-headerToPair cp _ (DeleteCookie key) =
+headerToPair (AddCookie sc) =
+    ("Set-Cookie", toByteString $ renderSetCookie $ sc)
+headerToPair (DeleteCookie key path) =
     ( "Set-Cookie"
-    , key `mappend` "=; path=" `mappend` cp `mappend` "; expires=Thu, 01-Jan-1970 00:00:00 GMT"
+    , S.concat
+        [ key
+        , "=; path="
+        , path
+        , "; expires=Thu, 01-Jan-1970 00:00:00 GMT"
+        ]
     )
-headerToPair _ _ (Header key value) = (key, value)
+headerToPair (Header key value) = (CI.mk key, value)
 
 -- | Get a unique identifier.
-newIdent :: MonadIO mo => GGHandler sub master mo String -- FIXME use Text
+newIdent :: GHandler sub master Text
 newIdent = do
     x <- get
     let i' = ghsIdent x + 1
     put x { ghsIdent = i' }
-    return $ 'h' : show i'
-
-liftIOHandler :: MonadIO mo
-              => GGHandler sub master IO a
-              -> GGHandler sub master mo a
-liftIOHandler (ReaderT m) = ReaderT $ \r -> liftIO $ m r
+    return $ T.pack $ 'h' : show i'
 
 -- | Redirect to a POST resource.
 --
@@ -838,8 +828,10 @@ liftIOHandler (ReaderT m) = ReaderT $ \r -> liftIO $ m r
 -- POST form, and some Javascript to automatically submit the form. This can be
 -- useful when you need to post a plain link somewhere that needs to cause
 -- changes on the server.
-redirectToPost :: MonadIO mo => Route master -> GGHandler sub master mo a
-redirectToPost dest = hamletToRepHtml
+redirectToPost :: RedirectUrl master url => url -> GHandler sub master a
+redirectToPost url = do
+    urlText <- toTextUrl url
+    hamletToRepHtml
 #if GHC7
             [hamlet|
 #else
@@ -851,7 +843,7 @@ redirectToPost dest = hamletToRepHtml
     <head>
         <title>Redirecting...
     <body onload="document.getElementById('form').submit()">
-        <form id="form" method="post" action="@{dest}">
+        <form id="form" method="post" action=#{urlText}>
             <noscript>
                 <p>Javascript has been disabled; please click on the button below to be redirected.
             <input type="submit" value="Continue">
@@ -859,36 +851,83 @@ redirectToPost dest = hamletToRepHtml
 
 -- | Converts the given Hamlet template into 'Content', which can be used in a
 -- Yesod 'Response'.
-hamletToContent :: Monad mo
-                => HtmlUrl (Route master) -> GGHandler sub master mo Content
+hamletToContent :: HtmlUrl (Route master) -> GHandler sub master Content
 hamletToContent h = do
     render <- getUrlRenderParams
     return $ toContent $ h render
 
 -- | Wraps the 'Content' generated by 'hamletToContent' in a 'RepHtml'.
-hamletToRepHtml :: Monad mo
-                => HtmlUrl (Route master) -> GGHandler sub master mo RepHtml
+hamletToRepHtml :: HtmlUrl (Route master) -> GHandler sub master RepHtml
 hamletToRepHtml = liftM RepHtml . hamletToContent
 
 -- | Get the request\'s 'W.Request' value.
-waiRequest :: Monad mo => GGHandler sub master mo W.Request
+waiRequest :: GHandler sub master W.Request
 waiRequest = reqWaiRequest `liftM` getRequest
 
-getMessageRender :: (Monad mo, RenderMessage master message) => GGHandler s master mo (message -> Text)
+getMessageRender :: RenderMessage master message => GHandler s master (message -> Text)
 getMessageRender = do
     m <- getYesod
     l <- reqLangs `liftM` getRequest
     return $ renderMessage m l
 
-cacheLookup :: MonadIO mo => CacheKey a -> GGHandler sub master mo (Maybe a)
+cacheLookup :: CacheKey a -> GHandler sub master (Maybe a)
 cacheLookup k = do
     gs <- get
     return $ Cache.lookup k $ ghsCache gs
 
-cacheInsert :: MonadIO mo => CacheKey a -> a -> GGHandler sub master mo ()
+cacheInsert :: CacheKey a -> a -> GHandler sub master ()
 cacheInsert k v = modify $ \gs ->
     gs { ghsCache = Cache.insert k v $ ghsCache gs }
 
-cacheDelete :: MonadIO mo => CacheKey a -> GGHandler sub master mo ()
+cacheDelete :: CacheKey a -> GHandler sub master ()
 cacheDelete k = modify $ \gs ->
     gs { ghsCache = Cache.delete k $ ghsCache gs }
+
+ask :: GHandler sub master (HandlerData sub master)
+ask = GHandler return
+
+local :: (HandlerData sub' master' -> HandlerData sub master)
+      -> GHandler sub master a
+      -> GHandler sub' master' a
+local f (GHandler x) = GHandler $ \r -> x $ f r
+
+-- | The standard @MonadTrans@ class only allows lifting for monad
+-- transformers. While @GHandler@ and @GWidget@ should allow lifting, their
+-- types do not express that they actually are transformers. This replacement
+-- class accounts for this.
+class MonadLift base m | m -> base where
+    lift :: base a -> m a
+instance (Monad m, MonadTrans t) => MonadLift m (t m) where
+    lift = Control.Monad.Trans.Class.lift
+instance MonadLift (ResourceT IO) (GHandler sub master) where
+    lift = GHandler . const
+
+-- Instances for GHandler
+instance Functor (GHandler sub master) where
+    fmap f (GHandler x) = GHandler $ \r -> fmap f (x r)
+instance Applicative (GHandler sub master) where
+    pure = GHandler . const . pure
+    GHandler f <*> GHandler x = GHandler $ \r -> f r <*> x r
+instance Monad (GHandler sub master) where
+    return = pure
+    GHandler x >>= f = GHandler $ \r -> x r >>= \x' -> unGHandler (f x') r
+instance MonadIO (GHandler sub master) where
+    liftIO = GHandler . const . lift
+instance MonadBase IO (GHandler sub master) where
+    liftBase = GHandler . const . lift
+instance MonadBaseControl IO (GHandler sub master) where
+    data StM (GHandler sub master) a = StH (StM (ResourceT IO) a)
+    liftBaseWith f = GHandler $ \reader ->
+        liftBaseWith $ \runInBase ->
+            f $ liftM StH . runInBase . (\(GHandler r) -> r reader)
+    restoreM (StH base) = GHandler $ const $ restoreM base
+
+instance Resource (GHandler sub master) where
+    type Base (GHandler sub master) = IO
+    resourceLiftBase = liftIO
+    resourceBracket_ a b c = control $ \run -> resourceBracket_ a b (run c)
+instance ResourceUnsafeIO (GHandler sub master) where
+    unsafeFromIO = liftIO
+instance ResourceThrow (GHandler sub master) where
+    resourceThrow = liftIO . throwIO
+instance ResourceIO (GHandler sub master)
