@@ -25,13 +25,21 @@ module Yesod.Internal.Core
     , formatLogMessage
     , fileLocationToString
     , messageLoggerHandler
+      -- * jsLoader
+    , ScriptLoadPosition (..)
+    , BottomOfHeadAsync
+    , loadJsYepnope
       -- * Misc
     , yesodVersion
     , yesodRender
+    , resolveApproot
+    , Approot (..)
     ) where
 
 import Yesod.Content
-import Yesod.Handler
+import Yesod.Handler hiding (lift, getExpires)
+
+import Yesod.Routes.Class
 
 import Control.Arrow ((***))
 import Control.Monad (forM)
@@ -46,7 +54,6 @@ import qualified Web.ClientSession as CS
 import qualified Data.ByteString.Char8 as S8
 import qualified Data.ByteString.Lazy as L
 import Data.Monoid
-import Control.Monad.Trans.Writer (runWriterT)
 import Text.Hamlet
 import Text.Julius
 import Text.Blaze ((!), customAttribute, textTag, toValue, unsafeLazyByteString)
@@ -59,7 +66,7 @@ import Web.Cookie (parseCookies)
 import qualified Data.Map as Map
 import Data.Time
 import Network.HTTP.Types (encodePath)
-import qualified Data.Text as TS
+import qualified Data.Text as T
 import Data.Text (Text)
 import qualified Data.Text.Encoding as TE
 import qualified Data.Text.Encoding.Error as TEE
@@ -67,6 +74,7 @@ import Blaze.ByteString.Builder (Builder, toByteString)
 import Blaze.ByteString.Builder.Char.Utf8 (fromText)
 import Data.List (foldl')
 import qualified Network.HTTP.Types as H
+import Web.Cookie (SetCookie (..))
 import qualified Data.Text.Lazy as TL
 import qualified Data.Text.Lazy.IO
 import qualified Data.Text.Lazy.Builder as TB
@@ -75,6 +83,7 @@ import Text.Blaze (preEscapedLazyText)
 import Data.Aeson (Value (Array, String))
 import Data.Aeson.Encode (encode)
 import qualified Data.Vector as Vector
+import Network.Wai.Middleware.Gzip (GzipSettings, def)
 
 -- mega repo can't access this
 #ifndef MEGA
@@ -84,7 +93,7 @@ yesodVersion :: String
 yesodVersion = showVersion Paths_yesod_core.version
 #else
 yesodVersion :: String
-yesodVersion = "0.9.3.2"
+yesodVersion = "0.9.4"
 #endif
 
 #if GHC7
@@ -93,42 +102,61 @@ yesodVersion = "0.9.3.2"
 #define HAMLET $hamlet
 #endif
 
-class Eq u => RenderRoute u where
-    renderRoute :: u -> ([Text], [(Text, Text)])
-
 -- | This class is automatically instantiated when you use the template haskell
 -- mkYesod function. You should never need to deal with it directly.
-class YesodDispatch a master where
+class YesodDispatch sub master where
     yesodDispatch
         :: Yesod master
-        => a
+        => master
+        -> sub
+        -> (Route sub -> Route master)
+        -> (Maybe CS.Key -> W.Application) -- ^ 404 handler
+        -> (Route sub -> Maybe CS.Key -> W.Application) -- ^ 405 handler
+        -> Text -- ^ request method
+        -> [Text] -- ^ pieces
         -> Maybe CS.Key
-        -> [Text]
-        -> master
-        -> (Route a -> Route master)
-        -> Maybe W.Application
+        -> W.Application
 
     yesodRunner :: Yesod master
-                => a
+                => GHandler sub master ChooseRep
                 -> master
-                -> (Route a -> Route master)
-                -> Maybe CS.Key -> Maybe (Route a) -> GHandler a master ChooseRep -> W.Application
+                -> sub
+                -> Maybe (Route sub)
+                -> (Route sub -> Route master)
+                -> Maybe CS.Key
+                -> W.Application
     yesodRunner = defaultYesodRunner
 
--- | Define settings for a Yesod applications. The only required setting is
--- 'approot'; other than that, there are intelligent defaults.
-class RenderRoute (Route a) => Yesod a where
+-- | How to determine the root of the application for constructing URLs.
+--
+-- Note that future versions of Yesod may add new constructors without bumping
+-- the major version number. As a result, you should /not/ pattern match on
+-- @Approot@ values.
+data Approot master = ApprootRelative -- ^ No application root.
+                    | ApprootStatic Text
+                    | ApprootMaster (master -> Text)
+                    | ApprootRequest (master -> W.Request -> Text)
+
+type ResolvedApproot = Text
+
+-- | Define settings for a Yesod applications. All methods have intelligent
+-- defaults, and therefore no implementation is required.
+class RenderRoute a => Yesod a where
     -- | An absolute URL to the root of the application. Do not include
     -- trailing slash.
     --
-    -- If you want to be lazy, you can supply an empty string under the
-    -- following conditions:
+    -- Default value: 'ApprootRelative'. This is valid under the following
+    -- conditions:
     --
     -- * Your application is served from the root of the domain.
     --
     -- * You do not use any features that require absolute URLs, such as Atom
     -- feeds and XML sitemaps.
-    approot :: a -> Text
+    --
+    -- If this is not true, you should override with a different
+    -- implementation.
+    approot :: Approot a
+    approot = ApprootRelative
 
     -- | The encryption key to be used for encrypting client sessions.
     -- Returning 'Nothing' disables sessions.
@@ -214,15 +242,15 @@ class RenderRoute (Route a) => Yesod a where
             then Right s
             else Left corrected
       where
-        corrected = filter (not . TS.null) s
+        corrected = filter (not . T.null) s
 
     -- | Builds an absolute URL by concatenating the application root with the  
     -- pieces of a path and a query string, if any. 
     -- Note that the pieces of the path have been previously cleaned up by 'cleanPath'.
     joinPath :: a
-             -> TS.Text -- ^ application root
-             -> [TS.Text] -- ^ path pieces
-             -> [(TS.Text, TS.Text)] -- ^ query string
+             -> T.Text -- ^ application root
+             -> [T.Text] -- ^ path pieces
+             -> [(T.Text, T.Text)] -- ^ query string
              -> Builder
     joinPath _ ar pieces' qs' = fromText ar `mappend` encodePath pieces qs
       where
@@ -247,10 +275,16 @@ class RenderRoute (Route a) => Yesod a where
                      -> GHandler sub a (Maybe (Either Text (Route a, [(Text, Text)])))
     addStaticContent _ _ _ = return Nothing
 
+    {- Temporarily disabled until we have a better interface.
     -- | Whether or not to tie a session to a specific IP address. Defaults to
-    -- 'True'.
+    -- 'False'.
+    --
+    -- Note: This setting has two known problems: it does not work correctly
+    -- when behind a reverse proxy (including load balancers), and it may not
+    -- function correctly if the user is behind a proxy.
     sessionIpAddress :: a -> Bool
-    sessionIpAddress _ = True
+    sessionIpAddress _ = False
+    -}
 
     -- | The path value to set for cookies. By default, uses \"\/\", meaning
     -- cookies will be sent to every page on the current domain.
@@ -261,27 +295,47 @@ class RenderRoute (Route a) => Yesod a where
     maximumContentLength :: a -> Maybe (Route a) -> Int
     maximumContentLength _ _ = 2 * 1024 * 1024 -- 2 megabytes
 
-    -- | Send a message to the log. By default, prints to stderr.
+    -- | Send a message to the log. By default, prints to stdout.
     messageLogger :: a
                   -> Loc -- ^ position in source code
                   -> LogLevel
                   -> Text -- ^ message
                   -> IO ()
-    messageLogger _ loc level msg =
-        formatLogMessage loc level msg >>=
-        Data.Text.Lazy.IO.putStrLn
+    messageLogger a loc level msg =
+        if level < logLevel a
+            then return ()
+            else
+                formatLogMessage loc level msg >>=
+                Data.Text.Lazy.IO.putStrLn
 
-    -- | Apply gzip compression to files. Default is false.
-    gzipCompressFiles :: a -> Bool
-    gzipCompressFiles _ = False
+    -- | The logging level in place for this application. Any messages below
+    -- this level will simply be ignored.
+    logLevel :: a -> LogLevel
+    logLevel _ = LevelInfo
 
-    -- | Location of yepnope.js, if any. If one is provided, then all
+    -- | GZIP settings.
+    gzipSettings :: a -> GzipSettings
+    gzipSettings _ = def
+
+    -- | Deprecated. Use 'jsloader'. To use yepnope: jsLoader = BottomOfHeadAsync (loadJsYepnope eyn)
+    -- Location of yepnope.js, if any. If one is provided, then all
     -- Javascript files will be loaded asynchronously.
     yepnopeJs :: a -> Maybe (Either Text (Route a))
     yepnopeJs _ = Nothing
 
-messageLoggerHandler :: (Yesod m, MonadIO mo)
-                     => Loc -> LogLevel -> Text -> GGHandler s m mo ()
+    -- | Where to Load sripts from. We recommend changing this to 'BottomOfBody'
+    -- Alternatively use the built in async yepnope loader:
+    --
+    -- > BottomOfHeadAsync $ loadJsYepnope $ Right $ StaticR js_modernizr_js
+    --
+    -- Or write your own async js loader: see 'loadJsYepnope'
+    jsLoader :: a -> ScriptLoadPosition a
+    jsLoader y = case yepnopeJs y of
+                   Nothing  -> BottomOfHeadBlocking
+                   Just eyn -> BottomOfHeadAsync (loadJsYepnope eyn)
+
+messageLoggerHandler :: Yesod m
+                     => Loc -> LogLevel -> Text -> GHandler s m ()
 messageLoggerHandler loc level msg = do
     y <- getYesod
     liftIO $ messageLogger y loc level msg
@@ -294,7 +348,7 @@ instance Lift LogLevel where
     lift LevelInfo = [|LevelInfo|]
     lift LevelWarn = [|LevelWarn|]
     lift LevelError = [|LevelError|]
-    lift (LevelOther x) = [|LevelOther $ TS.pack $(lift $ TS.unpack x)|]
+    lift (LevelOther x) = [|LevelOther $ T.pack $(lift $ T.unpack x)|]
 
 formatLogMessage :: Loc
                  -> LogLevel
@@ -303,13 +357,13 @@ formatLogMessage :: Loc
 formatLogMessage loc level msg = do
     now <- getCurrentTime
     return $ TB.toLazyText $
-        TB.fromText (TS.pack $ show now)
+        TB.fromText (T.pack $ show now)
         `mappend` TB.fromText " ["
-        `mappend` TB.fromText (TS.pack $ drop 5 $ show level)
+        `mappend` TB.fromText (T.pack $ drop 5 $ show level)
         `mappend` TB.fromText "] "
         `mappend` TB.fromText msg
         `mappend` TB.fromText " @("
-        `mappend` TB.fromText (TS.pack $ fileLocationToString loc)
+        `mappend` TB.fromText (T.pack $ fileLocationToString loc)
         `mappend` TB.fromText ") "
 
 -- taken from file-location package
@@ -323,15 +377,15 @@ fileLocationToString loc = (loc_package loc) ++ ':' : (loc_module loc) ++
     char = show . snd . loc_start
 
 defaultYesodRunner :: Yesod master
-                   => a
+                   => GHandler sub master ChooseRep
                    -> master
-                   -> (Route a -> Route master)
+                   -> sub
+                   -> Maybe (Route sub)
+                   -> (Route sub -> Route master)
                    -> Maybe CS.Key
-                   -> Maybe (Route a)
-                   -> GHandler a master ChooseRep
                    -> W.Application
-defaultYesodRunner _ m toMaster _ murl _ req
-    | maximumContentLength m (fmap toMaster murl) < len =
+defaultYesodRunner _ master _ murl toMaster _ req
+    | maximumContentLength master (fmap toMaster murl) < len =
         return $ W.responseLBS
             (H.Status 413 "Too Large")
             [("Content-Type", "text/plain")]
@@ -342,12 +396,12 @@ defaultYesodRunner _ m toMaster _ murl _ req
         case reads $ S8.unpack s of
             [] -> Nothing
             (x, _):_ -> Just x
-defaultYesodRunner s master toMasterRoute mkey murl handler req = do
+defaultYesodRunner handler master sub murl toMasterRoute mkey req = do
     now <- {-# SCC "getCurrentTime" #-} liftIO getCurrentTime
     let getExpires m = {-# SCC "getExpires" #-} fromIntegral (m * 60) `addUTCTime` now
     let exp' = {-# SCC "exp'" #-} getExpires $ clientSessionDuration master
-    let rh = {-# SCC "rh" #-} takeWhile (/= ':') $ show $ W.remoteHost req
-    let host = if sessionIpAddress master then S8.pack rh else ""
+    --let rh = {-# SCC "rh" #-} takeWhile (/= ':') $ show $ W.remoteHost req
+    let host = "" -- FIXME if sessionIpAddress master then S8.pack rh else ""
     let session' = {-# SCC "session'" #-}
             case mkey of
                 Nothing -> []
@@ -369,13 +423,14 @@ defaultYesodRunner s master toMasterRoute mkey murl handler req = do
                             Nothing ->
                                 permissionDenied "Authentication required"
                             Just url' -> do
-                                setUltDest'
-                                redirect RedirectTemporary url'
+                                setUltDestCurrent
+                                redirect url'
                     Unauthorized s' -> permissionDenied s'
                 handler
     let sessionMap = Map.fromList
                    $ filter (\(x, _) -> x /= nonceKey) session'
-    yar <- handlerToYAR master s toMasterRoute (yesodRender master) errorHandler rr murl sessionMap h
+    let ra = resolveApproot master req
+    yar <- handlerToYAR master sub toMasterRoute (yesodRender master ra) errorHandler rr murl sessionMap h
     let mnonce = reqNonce rr
     -- FIXME should we be caching this IV value and reusing it for efficiency?
     iv <- {-# SCC "iv" #-} maybe (return $ error "Should not be used") (const $ liftIO CS.randomIV) mkey
@@ -389,17 +444,21 @@ defaultYesodRunner s master toMasterRoute mkey murl handler req = do
                 (Just key, Just nonce)
                     -> encodeSession key iv exp' host
                      $ Map.toList
-                     $ Map.insert nonceKey nonce sm
+                     $ Map.insert nonceKey (TE.encodeUtf8 nonce) sm
                 _ -> mempty
         hs' =
             case mkey of
                 Nothing -> hs
-                Just _ -> AddCookie
-                            (clientSessionDuration master)
-                            sessionName
-                            sessionVal
+                Just _ -> AddCookie def
+                            { setCookieName = sessionName
+                            , setCookieValue = sessionVal
+                            , setCookiePath = Just (cookiePath master)
+                            , setCookieExpires = Just $ getExpires (clientSessionDuration master)
+                            , setCookieDomain = Nothing
+                            , setCookieHttpOnly = True
+                            }
                           : hs
-        hs'' = map (headerToPair (cookiePath master) getExpires) hs'
+        hs'' = map headerToPair hs'
         hs''' = ("Content-Type", ct) : hs''
 
 data AuthResult = Authorized | AuthenticationRequired | Unauthorized Text
@@ -499,12 +558,12 @@ jelper = fmap jsToHtml
 widgetToPageContent :: (Eq (Route master), Yesod master)
                     => GWidget sub master ()
                     -> GHandler sub master (PageContent (Route master))
-widgetToPageContent (GWidget w) = do
+widgetToPageContent w = do
     master <- getYesod
-    ((), GWData (Body body) (Last mTitle) scripts' stylesheets' style jscript (Head head')) <- runWriterT w
+    ((), GWData (Body body) (Last mTitle) scripts' stylesheets' style jscript (Head head')) <- unGWidget w
     let title = maybe mempty unTitle mTitle
-    let scripts = runUniqueList scripts'
-    let stylesheets = runUniqueList stylesheets'
+        scripts = runUniqueList scripts'
+        stylesheets = runUniqueList stylesheets'
 
     render <- getUrlRenderParams
     let renderLoc x =
@@ -528,22 +587,20 @@ widgetToPageContent (GWidget w) = do
                    $ encodeUtf8 $ renderJavascriptUrl render s
                 return $ renderLoc x
 
-    let addAttr x (y, z) = x ! customAttribute (textTag y) (toValue z)
-    let renderLoc' render' (Local url) = render' url []
-        renderLoc' _ (Remote s) = s
-    let mkScriptTag (Script loc attrs) render' =
-            foldl' addAttr TBH.script (("src", renderLoc' render' loc) : attrs) $ return ()
-    let mkLinkTag (Stylesheet loc attrs) render' =
-            foldl' addAttr TBH.link
-                ( ("rel", "stylesheet")
-                : ("href", renderLoc' render' loc)
-                : attrs
-                )
-    let left (Left x) = Just x
-        left _ = Nothing
-        right (Right x) = Just x
-        right _ = Nothing
-    let head'' = [HAMLET|
+    -- modernizr should be at the end of the <head> http://www.modernizr.com/docs/#installing
+    -- the asynchronous loader means your page doesn't have to wait for all the js to load
+    let (mcomplete, asyncScripts) = asyncHelper render scripts jscript jsLoc
+        regularScriptLoad = [HAMLET|
+$forall s <- scripts
+    ^{mkScriptTag s}
+$maybe j <- jscript
+    $maybe s <- jsLoc
+        <script src="#{s}">
+    $nothing
+        <script>^{jelper j}
+|]
+        headAll = [HAMLET|
+\^{head'}
 $forall s <- stylesheets
     ^{mkLinkTag s}
 $forall s <- css
@@ -557,39 +614,76 @@ $forall s <- css
             <style media=#{media}>#{content}
         $nothing
             <style>#{content}
-$maybe _ <- yepnopeJs master
-$nothing
-    $forall s <- scripts
-        ^{mkScriptTag s}
-    $maybe j <- jscript
-        $maybe s <- jsLoc
-            <script src="#{s}">
-        $nothing
-            <script>^{jelper j}
-\^{head'}
+$case jsLoader master
+  $of BottomOfHeadAsync asyncJsLoader
+      ^{asyncJsLoader asyncScripts mcomplete}
+  $of BottomOfHeadBlocking
+      ^{regularScriptLoad}
+  $of BottomOfBody
 |]
-    let (mcomplete, ynscripts) = ynHelper render scripts jscript jsLoc
-    let bodyYN = [HAMLET|
+    let bodyScript = [HAMLET|
 ^{body}
-$maybe eyn <- yepnopeJs master
+^{regularScriptLoad}
+|]
+    return $ PageContent title headAll (case jsLoader master of
+      BottomOfBody -> bodyScript
+      _ -> body)
+  where
+    renderLoc' render' (Local url) = render' url []
+    renderLoc' _ (Remote s) = s
+
+    addAttr x (y, z) = x ! customAttribute (textTag y) (toValue z)
+    mkScriptTag (Script loc attrs) render' =
+        foldl' addAttr TBH.script (("src", renderLoc' render' loc) : attrs) $ return ()
+    mkLinkTag (Stylesheet loc attrs) render' =
+        foldl' addAttr TBH.link
+            ( ("rel", "stylesheet")
+            : ("href", renderLoc' render' loc)
+            : attrs
+            )
+
+data ScriptLoadPosition master
+    = BottomOfBody
+    | BottomOfHeadBlocking
+    | BottomOfHeadAsync (BottomOfHeadAsync master)
+
+type BottomOfHeadAsync master
+       = [Text] -- ^ urls to load asynchronously
+      -> Maybe (HtmlUrl (Route master)) -- ^ widget of js to run on async completion
+      -> (HtmlUrl (Route master)) -- ^ widget to insert at the bottom of <head>
+
+left :: Either a b -> Maybe a
+left (Left x) = Just x
+left _ = Nothing
+
+right :: Either a b -> Maybe b
+right (Right x) = Just x
+right _ = Nothing
+
+jsonArray :: [Text] -> Html
+jsonArray = unsafeLazyByteString . encode . Array . Vector.fromList . map String
+
+-- | For use with setting 'jsLoader' to 'BottomOfHeadAsync'
+loadJsYepnope :: Yesod master => Either Text (Route master) -> [Text] -> Maybe (HtmlUrl (Route master)) -> (HtmlUrl (Route master))
+loadJsYepnope eyn scripts mcomplete =
+  [HAMLET|
     $maybe yn <- left eyn
         <script src=#{yn}>
     $maybe yn <- right eyn
         <script src=@{yn}>
     $maybe complete <- mcomplete
-        <script>yepnope({load:#{ynscripts},complete:function(){^{complete}}})
+        <script>yepnope({load:#{jsonArray scripts},complete:function(){^{complete}}});
     $nothing
-        <script>yepnope({load:#{ynscripts}})
+        <script>yepnope({load:#{jsonArray scripts}});
 |]
-    return $ PageContent title head'' bodyYN
 
-ynHelper :: (url -> [x] -> Text)
+asyncHelper :: (url -> [x] -> Text)
          -> [Script (url)]
          -> Maybe (JavascriptUrl (url))
          -> Maybe Text
-         -> (Maybe (HtmlUrl (url)), Html)
-ynHelper render scripts jscript jsLoc =
-    (mcomplete, unsafeLazyByteString $ encode $ Array $ Vector.fromList $ map String scripts'')
+         -> (Maybe (HtmlUrl url), [Text])
+asyncHelper render scripts jscript jsLoc =
+    (mcomplete, scripts'')
   where
     scripts' = map goScript scripts
     scripts'' =
@@ -608,14 +702,23 @@ ynHelper render scripts jscript jsLoc =
 
 yesodRender :: Yesod y
             => y
+            -> ResolvedApproot
             -> Route y
-            -> [(Text, Text)]
+            -> [(Text, Text)] -- ^ url query string
             -> Text
-yesodRender y u qs =
+yesodRender y ar url params =
     TE.decodeUtf8 $ toByteString $
     fromMaybe
-        (joinPath y (approot y) ps
-          $ qs ++ qs')
-        (urlRenderOverride y u)
+        (joinPath y ar ps
+          $ params ++ params')
+        (urlRenderOverride y url)
   where
-    (ps, qs') = renderRoute u
+    (ps, params') = renderRoute url
+
+resolveApproot :: Yesod master => master -> W.Request -> ResolvedApproot
+resolveApproot master req =
+    case approot of
+        ApprootRelative -> ""
+        ApprootStatic t -> t
+        ApprootMaster f -> f master
+        ApprootRequest f -> f master req
