@@ -9,6 +9,7 @@ import Test.Hspec.HUnit ()
 import Yesod.Core hiding (Request)
 import Network.Wai
 import Network.Wai.Test
+import Network.HTTP.Types (status301)
 
 data Y = Y
 mkYesod "Y" [parseRoutes|
@@ -17,7 +18,7 @@ mkYesod "Y" [parseRoutes|
 |]
 
 instance Yesod Y where
-    approot _ = "http://test"
+    approot = ApprootStatic "http://test"
     errorHandler (InternalError e) = return $ chooseRep $ RepPlain $ toContent e
     errorHandler x = defaultErrorHandler x
 
@@ -27,7 +28,7 @@ getRootR = error "FOOBAR" >> return ()
 getRedirR :: Handler ()
 getRedirR = do
     setHeader "foo" "bar"
-    redirect RedirectPermanent RootR
+    redirectWith status301 RootR
 
 exceptionsTest :: [Spec]
 exceptionsTest = describe "Test.Exceptions"
