@@ -18,8 +18,10 @@ randomStringSpecs = describe "Yesod.Internal.Request.randomString"
 
 -- NOTE: this testcase may break on other systems/architectures if
 -- mkStdGen is not identical everywhere (is it?).
+looksRandom :: Bool
 looksRandom = randomString 20 (mkStdGen 0) == "VH9SkhtptqPs6GqtofVg"
 
+noRepeat :: Int -> Int -> Bool
 noRepeat len n = length (nub $ map (randomString len . mkStdGen) [1..n]) == n
 
 
@@ -36,15 +38,19 @@ nonceSpecs = describe "Yesod.Internal.Request.parseWaiRequest (reqNonce)"
   , it "generates a new nonce for sessions without nonce" generateNonce
   ]
 
+noDisabledNonce :: Bool
 noDisabledNonce = reqNonce r == Nothing where
   r = parseWaiRequest' defaultRequest [] Nothing g
 
+ignoreDisabledNonce :: Bool
 ignoreDisabledNonce = reqNonce r == Nothing where
   r = parseWaiRequest' defaultRequest [("_NONCE", "old")] Nothing g
 
+useOldNonce :: Bool
 useOldNonce = reqNonce r == Just "old" where
   r = parseWaiRequest' defaultRequest [("_NONCE", "old")] (Just undefined) g
 
+generateNonce :: Bool
 generateNonce = reqNonce r /= Nothing where
   r = parseWaiRequest' defaultRequest [("_NONCE", "old")] (Just undefined) g
 
@@ -58,28 +64,33 @@ langSpecs = describe "Yesod.Internal.Request.parseWaiRequest (reqLangs)"
   , it "prioritizes correctly" prioritizeLangs
   ]
 
-respectAcceptLangs = reqLangs r == ["accept1", "accept2"] where
+respectAcceptLangs :: Bool
+respectAcceptLangs = reqLangs r == ["en-US", "es", "en"] where
   r = parseWaiRequest' defaultRequest
-        { requestHeaders = [("Accept-Language", "accept1, accept2")] } [] Nothing g
+        { requestHeaders = [("Accept-Language", "en-US, es")] } [] Nothing g
 
-respectSessionLang = reqLangs r == ["session"] where
-  r = parseWaiRequest' defaultRequest [("_LANG", "session")] Nothing g
+respectSessionLang :: Bool
+respectSessionLang = reqLangs r == ["en"] where
+  r = parseWaiRequest' defaultRequest [("_LANG", "en")] Nothing g
 
-respectCookieLang = reqLangs r == ["cookie"] where
+respectCookieLang :: Bool
+respectCookieLang = reqLangs r == ["en"] where
   r = parseWaiRequest' defaultRequest
-        { requestHeaders = [("Cookie", "_LANG=cookie")]
+        { requestHeaders = [("Cookie", "_LANG=en")]
         } [] Nothing g
 
-respectQueryLang = reqLangs r == ["query"] where
-  r = parseWaiRequest' defaultRequest { queryString = [("_LANG", Just "query")] } [] Nothing g
+respectQueryLang :: Bool
+respectQueryLang = reqLangs r == ["en-US", "en"] where
+  r = parseWaiRequest' defaultRequest { queryString = [("_LANG", Just "en-US")] } [] Nothing g
 
-prioritizeLangs = reqLangs r == ["query", "cookie", "session", "accept1", "accept2"] where
+prioritizeLangs :: Bool
+prioritizeLangs = reqLangs r == ["en-QUERY", "en-COOKIE", "en-SESSION", "en", "es"] where
   r = parseWaiRequest' defaultRequest
-        { requestHeaders = [ ("Accept-Language", "accept1, accept2")
-                           , ("Cookie", "_LANG=cookie")
+        { requestHeaders = [ ("Accept-Language", "en, es")
+                           , ("Cookie", "_LANG=en-COOKIE")
                            ]
-        , queryString = [("_LANG", Just "query")]
-        } [("_LANG", "session")] Nothing g
+        , queryString = [("_LANG", Just "en-QUERY")]
+        } [("_LANG", "en-SESSION")] Nothing g
 
 
 internalRequestTest :: [Spec]
@@ -87,5 +98,3 @@ internalRequestTest = descriptions [ randomStringSpecs
                                    , nonceSpecs
                                    , langSpecs
                                    ]
-
-main = hspec internalRequestTest
